import copy
from datetime import datetime

import dask.array as da
import numpy as np
import pandas as pd
import pytest
import xarray as xr

from act.io.arm import read_arm_netcdf
from act.qc.arm import add_dqr_to_qc
from act.qc.qcfilter import parse_bit, set_bit, unset_bit
from act.tests import EXAMPLE_MET1, EXAMPLE_METE40, EXAMPLE_IRT25m20s

try:
    import scikit_posthocs  # noqa

    SCIKIT_POSTHOCS_AVAILABLE = True
except ImportError:
    SCIKIT_POSTHOCS_AVAILABLE = False


def test_qc_test_errors():
    ds = read_arm_netcdf(EXAMPLE_MET1)
    var_name = 'temp_mean'

    assert ds.qcfilter.add_less_test(var_name, None) is None
    assert ds.qcfilter.add_greater_test(var_name, None) is None
    assert ds.qcfilter.add_less_equal_test(var_name, None) is None
    assert ds.qcfilter.add_equal_to_test(var_name, None) is None
    assert ds.qcfilter.add_not_equal_to_test(var_name, None) is None


def test_arm_qc():
    # Test DQR Webservice using known DQR
    variable = 'wspd_vec_mean'
    ds = read_arm_netcdf(EXAMPLE_METE40)
    ds_org = copy.deepcopy(ds)
    qc_variable = ds.qcfilter.check_for_ancillary_qc(variable)

    # DQR webservice does go down, so ensure it properly runs first before testing
    try:
        ds = add_dqr_to_qc(ds)

    except ValueError:
        return

    assert 'Suspect' not in ds[qc_variable].attrs['flag_assessments']
    assert 'Incorrect' not in ds[qc_variable].attrs['flag_assessments']
    assert 'Bad' in ds[qc_variable].attrs['flag_assessments']
    assert 'Indeterminate' in ds[qc_variable].attrs['flag_assessments']

    # Check that defualt will update all variables in DQR
    for var_name in ['wdir_vec_mean', 'wdir_vec_std', 'wspd_arith_mean', 'wspd_vec_mean']:
        qc_var = ds.qcfilter.check_for_ancillary_qc(var_name)
        assert ds[qc_var].attrs['flag_meanings'][-1].startswith('D190529.4')

    # Check that variable keyword works as expected.
    ds = copy.deepcopy(ds_org)
    add_dqr_to_qc(ds, variable=variable)
    qc_var = ds.qcfilter.check_for_ancillary_qc(variable)
    assert ds[qc_var].attrs['flag_meanings'][-1].startswith('D190529.4')
    qc_var = ds.qcfilter.check_for_ancillary_qc('wdir_vec_std')
    assert len(ds[qc_var].attrs['flag_masks']) == 0

    # Check that include and exclude keywords work as expected
    ds = copy.deepcopy(ds_org)
    add_dqr_to_qc(ds, variable=variable, exclude=['D190529.4'])
    assert len(ds[qc_variable].attrs['flag_meanings']) == 4
    add_dqr_to_qc(ds, variable=variable, include=['D400101.1'])
    assert len(ds[qc_variable].attrs['flag_meanings']) == 4
    add_dqr_to_qc(ds, variable=variable, include=['D190529.4'])
    assert len(ds[qc_variable].attrs['flag_meanings']) == 5
    add_dqr_to_qc(ds, variable=variable, assessment='Incorrect')
    assert len(ds[qc_variable].attrs['flag_meanings']) == 5

    # Test additional keywords
    add_dqr_to_qc(
        ds,
        variable=variable,
        assessment='Suspect',
        cleanup_qc=False,
        dqr_link=True,
        skip_location_vars=True,
    )
    assert len(ds[qc_variable].attrs['flag_meanings']) == 6

    # Default is to normalize assessment terms. Check that we can turn off.
    add_dqr_to_qc(ds, variable=variable, normalize_assessment=False)
    assert 'Suspect' in ds[qc_variable].attrs['flag_assessments']

    # Test that an error is raised when no datastream global attributes
    with np.testing.assert_raises(ValueError):
        ds4 = copy.deepcopy(ds)
        del ds4.attrs['datastream']
        del ds4.attrs['_datastream']
        add_dqr_to_qc(ds4, variable=variable)


def test_qcfilter():
    ds = read_arm_netcdf(EXAMPLE_IRT25m20s)
    var_name = 'inst_up_long_dome_resist'
    expected_qc_var_name = 'qc_' + var_name

    ds.qcfilter.check_for_ancillary_qc(
        var_name, add_if_missing=True, cleanup=False, flag_type=False
    )
    assert expected_qc_var_name in list(ds.keys())
    del ds[expected_qc_var_name]

    # Perform adding of quality control variables to Xarray dataset
    result = ds.qcfilter.add_test(var_name, test_meaning='Birds!')
    assert isinstance(result, dict)
    qc_var_name = result['qc_variable_name']
    assert qc_var_name == expected_qc_var_name

    # Check that new linking and describing attributes are set
    assert ds[qc_var_name].attrs['standard_name'] == 'quality_flag'
    assert ds[var_name].attrs['ancillary_variables'] == qc_var_name

    # Check that CF attributes are set including new flag_assessments
    assert 'flag_masks' in ds[qc_var_name].attrs.keys()
    assert 'flag_meanings' in ds[qc_var_name].attrs.keys()
    assert 'flag_assessments' in ds[qc_var_name].attrs.keys()

    # Check that the values of the attributes are set correctly
    assert ds[qc_var_name].attrs['flag_assessments'][0] == 'Bad'
    assert ds[qc_var_name].attrs['flag_meanings'][0] == 'Birds!'
    assert ds[qc_var_name].attrs['flag_masks'][0] == 1

    # Set some test values
    index = [0, 1, 2, 30]
    ds.qcfilter.set_test(var_name, index=index, test_number=result['test_number'])

    # Add a new test and set values
    index2 = [6, 7, 8, 50]
    ds.qcfilter.add_test(
        var_name,
        index=index2,
        test_number=9,
        test_meaning='testing high number',
        test_assessment='Suspect',
    )

    # Retrieve data from Xarray dataset as numpy masked array. Count number of masked
    # elements and ensure equal to size of index array.
    data = ds.qcfilter.get_masked_data(var_name, rm_assessments='Bad')
    assert np.ma.count_masked(data) == len(index)

    data = ds.qcfilter.get_masked_data(var_name, rm_assessments='Suspect', return_nan_array=True)
    assert np.sum(np.isnan(data)) == len(index2)

    data = ds.qcfilter.get_masked_data(
        var_name, rm_assessments=['Bad', 'Suspect'], ma_fill_value=np.nan
    )
    assert np.ma.count_masked(data) == len(index + index2)

    # Test internal function for returning the index array of where the
    # tests are set.
    assert (
        np.sum(
            ds.qcfilter.get_qc_test_mask(var_name, result['test_number'], return_index=True)
            - np.array(index, dtype=int)
        )
        == 0
    )

    # Test adding QC for length-1 variables
    ds['west'] = ('west', ['W'])
    ds['avg_wind_speed'] = ('west', [20])

    # Should not fail the test
    ds.qcfilter.add_test(
        'avg_wind_speed',
        index=ds.avg_wind_speed.data > 100,
        test_meaning='testing bool flag: false',
        test_assessment='Suspect',
    )
    assert ds.qc_avg_wind_speed.data == 0

    # Should fail the test
    ds.qcfilter.add_test(
        'avg_wind_speed',
        index=ds.avg_wind_speed.data < 100,
        test_meaning='testing bool flag: true',
        test_assessment='Suspect',
    )
    assert ds.qc_avg_wind_speed.data == 2

    # Should fail the test
    ds.qcfilter.add_test(
        'avg_wind_speed',
        index=[0],
        test_meaning='testing idx flag: true',
        test_assessment='Suspect',
    )
    assert ds.qc_avg_wind_speed.data == 6

    # Should not fail the test
    ds.qcfilter.add_test(
        'avg_wind_speed',
        test_meaning='testing idx flag: false',
        test_assessment='Suspect',
    )
    assert ds.qc_avg_wind_speed.data == 6

    # Unset a test
    ds.qcfilter.unset_test(var_name, index=0, test_number=result['test_number'])
    # Remove the test
    ds.qcfilter.remove_test(var_name, test_number=33)

    # Ensure removal works when flag_masks is a numpy array
    ds['qc_' + var_name].attrs['flag_masks'] = np.array(ds['qc_' + var_name].attrs['flag_masks'])
    ds.qcfilter.remove_test(var_name, test_number=result['test_number'])
    pytest.raises(ValueError, ds.qcfilter.add_test, var_name)
    pytest.raises(ValueError, ds.qcfilter.remove_test, var_name)

    ds.close()

    assert np.all(parse_bit([257]) == np.array([1, 9], dtype=np.int32))
    pytest.raises(ValueError, parse_bit, [1, 2])
    pytest.raises(ValueError, parse_bit, -1)

    assert set_bit(0, 16) == 32768
    data = range(0, 4)
    assert isinstance(set_bit(list(data), 2), list)
    assert isinstance(set_bit(tuple(data), 2), tuple)
    assert isinstance(unset_bit(list(data), 2), list)
    assert isinstance(unset_bit(tuple(data), 2), tuple)

    # Fill in missing tests
    ds = read_arm_netcdf(EXAMPLE_IRT25m20s)
    del ds[var_name].attrs['long_name']
    # Test creating a qc variable
    ds.qcfilter.create_qc_variable(var_name)
    # Test creating a second qc variable and of flag type
    ds.qcfilter.create_qc_variable(var_name, flag_type=True)
    result = ds.qcfilter.add_test(
        var_name,
        index=[1, 2, 3],
        test_number=9,
        test_meaning='testing high number',
        flag_value=True,
    )
    ds.qcfilter.set_test(var_name, index=5, test_number=9, flag_value=True)
    data = ds.qcfilter.get_masked_data(var_name)
    assert np.isclose(np.sum(data), 42674.766, 0.01)
    data = ds.qcfilter.get_masked_data(var_name, rm_assessments='Bad')
    assert np.isclose(np.sum(data), 42643.195, 0.01)

    ds.qcfilter.unset_test(var_name, test_number=9, flag_value=True)
    ds.qcfilter.unset_test(var_name, index=1, test_number=9, flag_value=True)
    assert ds.qcfilter.available_bit(result['qc_variable_name']) == 10
    assert ds.qcfilter.available_bit(result['qc_variable_name'], recycle=True) == 1
    ds.qcfilter.remove_test(var_name, test_number=9, flag_value=True)

    ds.qcfilter.update_ancillary_variable(var_name)
    # Test updating ancillary variable if does not exist
    ds.qcfilter.update_ancillary_variable('not_a_variable_name')
    # Change ancillary_variables attribute to test if add correct qc variable correctly
    ds[var_name].attrs['ancillary_variables'] = 'a_different_name'
    ds.qcfilter.update_ancillary_variable(var_name, qc_var_name=expected_qc_var_name)
    assert expected_qc_var_name in ds[var_name].attrs['ancillary_variables']

    # Test flag QC
    var_name = 'inst_sfc_ir_temp'
    qc_var_name = 'qc_' + var_name
    ds.qcfilter.create_qc_variable(var_name, flag_type=True)
    assert qc_var_name in list(ds.data_vars)
    assert 'flag_values' in ds[qc_var_name].attrs.keys()
    assert 'flag_masks' not in ds[qc_var_name].attrs.keys()
    del ds[qc_var_name]

    qc_var_name = ds.qcfilter.check_for_ancillary_qc(
        var_name, add_if_missing=True, cleanup=False, flag_type=True
    )
    assert qc_var_name in list(ds.data_vars)
    assert 'flag_values' in ds[qc_var_name].attrs.keys()
    assert 'flag_masks' not in ds[qc_var_name].attrs.keys()
    del ds[qc_var_name]

    ds.qcfilter.add_missing_value_test(var_name, flag_value=True, prepend_text='arm')
    ds.qcfilter.add_test(
        var_name,
        index=list(range(0, 20)),
        test_number=2,
        test_meaning='Testing flag',
        flag_value=True,
        test_assessment='Suspect',
    )
    assert qc_var_name in list(ds.data_vars)
    assert 'flag_values' in ds[qc_var_name].attrs.keys()
    assert 'flag_masks' not in ds[qc_var_name].attrs.keys()
    assert 'standard_name' in ds[qc_var_name].attrs.keys()
    assert ds[qc_var_name].attrs['flag_values'] == [1, 2]
    assert ds[qc_var_name].attrs['flag_assessments'] == ['Bad', 'Suspect']

    ds.close()


@pytest.mark.skipif(not SCIKIT_POSTHOCS_AVAILABLE, reason='scikit_posthocs is not installed.')
def test_qcfilter2():
    ds = read_arm_netcdf(EXAMPLE_IRT25m20s)
    var_name = 'inst_up_long_dome_resist'
    expected_qc_var_name = 'qc_' + var_name

    data = ds[var_name].values
    data[0:4] = data[0:4] + 30.0
    data[1000:1024] = data[1000:1024] + 30.0
    ds[var_name].values = data

    coef = 1.4
    ds.qcfilter.add_iqr_test(var_name, coef=1.4, test_assessment='Bad', prepend_text='arm')
    assert np.sum(ds[expected_qc_var_name].values) == 28
    assert ds[expected_qc_var_name].attrs['flag_masks'] == [1]
    assert ds[expected_qc_var_name].attrs['flag_meanings'] == [
        f'arm: Value outside of interquartile range test range with a coefficient of {coef}'
    ]

    ds.qcfilter.add_iqr_test(var_name, test_number=3, prepend_text='ACT')
    assert np.sum(ds[expected_qc_var_name].values) == 140
    assert ds[expected_qc_var_name].attrs['flag_masks'] == [1, 4]
    assert ds[expected_qc_var_name].attrs['flag_meanings'][-1] == (
        'ACT: Value outside of interquartile range test range with a coefficient of 1.5'
    )

    ds.qcfilter.add_gesd_test(var_name, test_assessment='Bad')
    assert np.sum(ds[expected_qc_var_name].values) == 204
    assert ds[expected_qc_var_name].attrs['flag_masks'] == [1, 4, 8]
    assert ds[expected_qc_var_name].attrs['flag_meanings'][-1] == (
        'Value failed generalized Extreme Studentized Deviate test with an alpha of 0.05'
    )

    ds.qcfilter.add_gesd_test(var_name, alpha=0.1)
    assert np.sum(ds[expected_qc_var_name].values) == 332
    assert ds[expected_qc_var_name].attrs['flag_masks'] == [1, 4, 8, 16]
    assert ds[expected_qc_var_name].attrs['flag_meanings'][-1] == (
        'Value failed generalized Extreme Studentized Deviate test with an alpha of 0.1'
    )
    assert ds[expected_qc_var_name].attrs['flag_assessments'] == [
        'Bad',
        'Indeterminate',
        'Bad',
        'Indeterminate',
    ]


def test_qcfilter3():
    ds = read_arm_netcdf(EXAMPLE_IRT25m20s)
    var_name = 'inst_up_long_dome_resist'
    result = ds.qcfilter.add_test(var_name, index=range(0, 100), test_meaning='testing')
    qc_var_name = result['qc_variable_name']
    assert ds[qc_var_name].values.dtype.kind in np.typecodes['AllInteger']

    ds[qc_var_name].values = ds[qc_var_name].values.astype(np.float32)
    assert ds[qc_var_name].values.dtype.kind not in np.typecodes['AllInteger']

    result = ds.qcfilter.get_qc_test_mask(var_name=var_name, test_number=1, return_index=False)
    assert np.sum(result) == 100
    result = ds.qcfilter.get_qc_test_mask(var_name=var_name, test_number=1, return_index=True)
    assert np.sum(result) == 4950

    # Test where QC variables are not integer type
    ds = ds.resample(time='5min').mean(keep_attrs=True)
    ds.qcfilter.add_test(var_name, index=range(0, ds.time.size), test_meaning='Testing float')
    assert np.sum(ds[qc_var_name].values) == 582

    ds[qc_var_name].values = ds[qc_var_name].values.astype(np.float32)
    ds.qcfilter.remove_test(var_name, test_number=2)
    assert np.sum(ds[qc_var_name].values) == 6


def test_qc_speed():
    """
    This tests the speed of the QC module to ensure changes do not significantly
    slow down the module's processing.
    """

    n_variables = 100
    n_samples = 100

    time = pd.date_range(start='2022-02-17 00:00:00', end='2022-02-18 00:00:00', periods=n_samples)

    # Create data variables with random noise
    np.random.seed(42)
    noisy_data_mapping = {f'data_var_{i}': np.random.random(time.shape) for i in range(n_variables)}

    ds = xr.Dataset(
        data_vars={name: ('time', data) for name, data in noisy_data_mapping.items()},
        coords={'time': time},
    )

    start = datetime.utcnow()
    for name, var in noisy_data_mapping.items():
        failed_qc = var > 0.75  # Consider data above 0.75 as bad. Negligible time here.
        ds.qcfilter.add_test(name, index=failed_qc, test_meaning='Value above threshold')

    time_diff = datetime.utcnow() - start
    assert time_diff.seconds <= 4


def test_datafilter():
    ds = read_arm_netcdf(EXAMPLE_MET1, drop_variables=['base_time', 'time_offset'])
    ds.clean.cleanup()

    data_var_names = list(ds.data_vars)
    qc_var_names = [var_name for var_name in ds.data_vars if var_name.startswith('qc_')]
    data_var_names = list(set(data_var_names) - set(qc_var_names))
    data_var_names.sort()
    qc_var_names.sort()

    var_name = 'atmos_pressure'

    ds_1 = ds.mean()

    ds.qcfilter.add_less_test(var_name, 99, test_assessment='Bad')
    ds_filtered = copy.deepcopy(ds)
    ds_filtered.qcfilter.datafilter(rm_assessments='Bad')
    ds_2 = ds_filtered.mean()
    assert np.isclose(ds_1[var_name].values, 98.86, atol=0.01)
    assert np.isclose(ds_2[var_name].values, 99.15, atol=0.01)
    assert isinstance(ds_1[var_name].data, da.core.Array)
    assert 'act.qc.datafilter' in ds_filtered[var_name].attrs['history']

    ds_filtered = copy.deepcopy(ds)
    ds_filtered.qcfilter.datafilter(rm_assessments='Bad', variables=var_name, del_qc_var=True)
    ds_2 = ds_filtered.mean()
    assert np.isclose(ds_2[var_name].values, 99.15, atol=0.01)
    expected_var_names = sorted(list(set(data_var_names + qc_var_names) - {'qc_' + var_name}))
    assert sorted(list(ds_filtered.data_vars)) == expected_var_names

    ds_filtered = copy.deepcopy(ds)
    ds_filtered.qcfilter.datafilter(rm_assessments='Bad', del_qc_var=True)
    assert sorted(list(ds_filtered.data_vars)) == data_var_names

    ds.close()
    del ds


def test_qc_data_type():
    drop_vars = [
        'base_time',
        'time_offset',
        'inst_up_long_case_resist',
        'inst_up_long_hemisp_tp',
        'inst_up_short_hemisp_tp',
        'inst_sfc_ir_temp',
        'lat',
        'lon',
        'alt',
    ]
    ds = read_arm_netcdf(EXAMPLE_IRT25m20s, drop_variables=drop_vars)
    var_name = 'inst_up_long_dome_resist'
    expected_qc_var_name = 'qc_' + var_name
    ds.qcfilter.check_for_ancillary_qc(var_name, add_if_missing=True)
    del ds[expected_qc_var_name].attrs['flag_meanings']
    del ds[expected_qc_var_name].attrs['flag_assessments']
    ds[expected_qc_var_name] = ds[expected_qc_var_name].astype(np.int8)
    ds.qcfilter.add_test(var_name, index=[1], test_number=9, test_meaning='First test')

    assert ds[expected_qc_var_name].attrs['flag_masks'][0].dtype == np.uint32
    assert ds[expected_qc_var_name].dtype == np.int16
    ds.qcfilter.add_test(var_name, index=[1], test_number=17, test_meaning='Second test')
    assert ds[expected_qc_var_name].dtype == np.int32
    ds.qcfilter.add_test(var_name, index=[1], test_number=33, test_meaning='Third test')
    assert ds[expected_qc_var_name].dtype == np.int64
    assert ds[expected_qc_var_name].attrs['flag_masks'][0].dtype == np.uint64

    ds.qcfilter.add_test(var_name, index=[1], test_meaning='Fourth test', recycle=True)



import act


def test_croptype():
    year = 2018
    lat = 37.15
    lon = -98.362
    # Try for when the cropscape API is not working
    try:
        crop = act.discovery.cropscape.get_crop_type(lat, lon, year)
        crop2 = act.discovery.cropscape.get_crop_type(lat, lon)
    except Exception:
        return

    # print(crop, crop2)
    if crop is not None:
        assert crop == 'Dbl Crop WinWht/Sorghum'
    if crop2 is not None:
        # assert crop2 == 'Sorghum'
        assert crop2 in ['Soybeans', 'Winter Wheat']



import numpy as np
import pytest

import act

try:
    import pysp2  # noqa

    PYSP2_AVAILABLE = True
except ImportError:
    PYSP2_AVAILABLE = False


@pytest.mark.skipif(not PYSP2_AVAILABLE, reason='PySP2 is not installed.')
def test_sp2_waveform_stats():
    my_sp2b = act.io.read_sp2(act.tests.EXAMPLE_SP2B)
    my_ini = act.tests.EXAMPLE_INI
    my_binary = act.qc.get_waveform_statistics(my_sp2b, my_ini, parallel=False)
    assert my_binary.PkHt_ch1.max() == 62669.4
    np.testing.assert_almost_equal(np.nanmax(my_binary.PkHt_ch0.values), 98708.92915295, decimal=1)
    np.testing.assert_almost_equal(np.nanmax(my_binary.PkHt_ch4.values), 65088.39598033, decimal=1)


@pytest.mark.skipif(not PYSP2_AVAILABLE, reason='PySP2 is not installed.')
def test_sp2_psds():
    my_sp2b = act.io.read_sp2(act.tests.EXAMPLE_SP2B)
    my_ini = act.tests.EXAMPLE_INI
    my_binary = act.qc.get_waveform_statistics(my_sp2b, my_ini, parallel=False)
    my_hk = act.io.read_hk_file(act.tests.EXAMPLE_HK)
    my_binary = act.retrievals.calc_sp2_diams_masses(my_binary)
    scatrejectkey = my_binary['ScatRejectKey'].values
    assert np.nanmax(my_binary['ScatDiaBC50'].values[scatrejectkey == 0]) < 1000.0
    my_psds = act.retrievals.process_sp2_psds(my_binary, my_hk, my_ini)
    np.testing.assert_almost_equal(my_psds['NumConcIncan'].max(), 0.95805343)



"""
This module contains various retrievals for datasets.

"""

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=['aeri', 'cbh', 'doppler_lidar', 'irt', 'radiation', 'sonde', 'sp2'],
    submod_attrs={
        'aeri': ['aeri2irt'],
        'cbh': ['generic_sobel_cbh'],
        'doppler_lidar': ['compute_winds_from_ppi'],
        'irt': ['sst_from_irt', 'sum_function_irt'],
        'radiation': [
            'calculate_dsh_from_dsdh_sdn',
            'calculate_irradiance_stats',
            'calculate_longwave_radiation',
            'calculate_net_radiation',
        ],
        'sonde': [
            'calculate_pbl_liu_liang',
            'calculate_precipitable_water',
            'calculate_stability_indicies',
            'calculate_pbl_heffter',
        ],
        'sp2': ['calc_sp2_diams_masses', 'process_sp2_psds'],
    },
)



"""
This module contains classes for displaying data.
:func:`act.plotting.Display` is the
base class on which all other Display classes are inherited from. If you are making
a new Display object, please make it inherited from this class.

| :func:`act.plotting.ContourDisplay` handles the plotting of contour plots.
| :func:`act.plotting.DistributionDisplay` handles the plotting of distribution-related plots.
| :func:`act.plotting.GeographicPlotDisplay` handles the plotting of lat-lon plots.
| :func:`act.plotting.SkewTDisplay` handles the plotting of Skew-T diagrams.
| :func:`act.plotting.TimeSeriesDisplay` handles the plotting of timeseries.
| :func:`act.plotting.WindRoseDisplay` handles the plotting of wind rose plots.
| :func:`act.plotting.XSectionDisplay` handles the plotting of cross sections.

"""

import lazy_loader as lazy

# Load colormaps
import cmweather  # noqa

# Eagerly load in common
from . import common  # noqa

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=[
        'act_cmap',
        '_act_cmap',
        'common',
        'contourdisplay',
        'geodisplay',
        'plot',
        'skewtdisplay',
        'timeseriesdisplay',
        'windrosedisplay',
        'xsectiondisplay',
        'distributiondisplay',
    ],
    submod_attrs={
        'contourdisplay': ['ContourDisplay'],
        'geodisplay': ['GeographicPlotDisplay'],
        'plot': ['Display', 'GroupByDisplay'],
        'skewtdisplay': ['SkewTDisplay'],
        'timeseriesdisplay': ['TimeSeriesDisplay'],
        'windrosedisplay': ['WindRoseDisplay'],
        'xsectiondisplay': ['XSectionDisplay'],
        'distributiondisplay': ['DistributionDisplay'],
    },
)



import numpy as np

import act


def test_aeri2irt():
    aeri_ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_AERI)
    aeri_ds = act.retrievals.aeri.aeri2irt(aeri_ds)
    assert np.round(np.nansum(aeri_ds['aeri_irt_equiv_temperature'].values)).astype(int) == 17372
    np.testing.assert_almost_equal(
        aeri_ds['aeri_irt_equiv_temperature'].values[7], 286.081, decimal=3
    )
    np.testing.assert_almost_equal(
        aeri_ds['aeri_irt_equiv_temperature'].values[-10], 285.366, decimal=3
    )
    aeri_ds.close()
    del aeri_ds



import xarray as xr
import numpy as np
import pandas as pd

from datetime import datetime


def read_hysplit(filename, base_year=2000):
    """
    Reads an input HYSPLIT trajectory for plotting in ACT.

    Parameters
    ----------
    filename: str
        The input file name.
    base_year: int
        The first year of the century in which the data are contained.

    Returns
    -------
    ds: xarray Dataset
        The ACT dataset containing the HYSPLIT trajectories
    """

    ds = xr.Dataset({})
    num_lines = 0
    with open(filename) as filebuf:
        num_grids = int(filebuf.readline().split()[0])
        num_lines += 1
        grid_times = []
        grid_names = []
        forecast_hours = np.zeros(num_grids)
        for i in range(num_grids):
            data = filebuf.readline().split()
            num_lines += 1
            grid_names.append(data[0])
            grid_times.append(
                datetime(year=int(data[1]), month=int(data[2]), day=int(data[3]), hour=int(data[4]))
            )
            forecast_hours[i] = int(data[5])
        ds["grid_forecast_hour"] = xr.DataArray(forecast_hours, dims=["num_grids"])
        ds["grid_forecast_hour"].attrs["standard_name"] = "Grid forecast hour"
        ds["grid_forecast_hour"].attrs["units"] = "Hour [UTC]"
        ds["grid_times"] = xr.DataArray(np.array(grid_times), dims=["num_grids"])
        data_line = filebuf.readline().split()
        num_lines += 1
        ds.attrs["trajectory_direction"] = data_line[1]
        ds.attrs["vertical_motion_calculation_method"] = data_line[2]
        num_traj = int(data_line[0])
        traj_times = []
        start_lats = np.zeros(num_traj)
        start_lons = np.zeros(num_traj)
        start_alt = np.zeros(num_traj)
        for i in range(num_traj):
            data = filebuf.readline().split()
            num_lines += 1
            traj_times.append(
                datetime(
                    year=(base_year + int(data[0])),
                    month=int(data[1]),
                    day=int(data[2]),
                    hour=int(data[3]),
                )
            )
            start_lats[i] = float(data[4])
            start_lons[i] = float(data[5])
            start_alt[i] = float(data[6])

        ds["start_latitude"] = xr.DataArray(start_lats, dims=["num_trajectories"])
        ds["start_latitude"].attrs["long_name"] = "Trajectory start latitude"
        ds["start_latitude"].attrs["units"] = "degree"
        ds["start_longitude"] = xr.DataArray(start_lats, dims=["num_trajectories"])
        ds["start_longitude"].attrs["long_name"] = "Trajectory start longitude"
        ds["start_longitude"].attrs["units"] = "degree"
        ds["start_altitude"] = xr.DataArray(start_alt, dims=["num_trajectories"])
        ds["start_altitude"].attrs["long_name"] = "Trajectory start altitude"
        ds["start_altitude"].attrs["units"] = "degree"
        data = filebuf.readline().split()
        num_lines += 1
        var_list = [
            "trajectory_number",
            "grid_number",
            "year",
            "month",
            "day",
            "hour",
            "minute",
            "forecast_hour",
            "age",
            "lat",
            "lon",
            "alt",
        ]
        for variable in data[1:]:
            var_list.append(variable)

    input_df = pd.read_csv(
        filename, sep=r'\s+', index_col=False, names=var_list, skiprows=12
    )  # noqa W605
    input_df['year'] = base_year + input_df['year']
    input_df['time'] = pd.to_datetime(
        input_df[["year", "month", "day", "hour", "minute"]], format='%y%m%d%H%M'
    )
    input_df = input_df.set_index("time")
    del input_df["year"]
    del input_df["month"]
    del input_df["day"]
    del input_df["hour"]
    del input_df["minute"]
    ds = ds.merge(input_df.to_xarray())
    ds.attrs['datastream'] = 'hysplit'
    ds["trajectory_number"].attrs["standard_name"] = "Trajectory number"
    ds["trajectory_number"].attrs["units"] = "1"
    ds["grid_number"].attrs["standard_name"] = "Grid number"
    ds["grid_number"].attrs["units"] = "1"
    ds["age"].attrs["standard_name"] = "Grid number"
    ds["age"].attrs["units"] = "1"
    ds["lat"].attrs["standard_name"] = "Latitude"
    ds["lat"].attrs["units"] = "degree"
    ds["lon"].attrs["standard_name"] = "Longitude"
    ds["lon"].attrs["units"] = "degree"
    ds["alt"].attrs["standard_name"] = "Altitude"
    ds["alt"].attrs["units"] = "meter"
    return ds



import act

from act.tests import sample_files


def test_read_hysplit():
    filename = sample_files.EXAMPLE_HYSPLIT
    ds = act.io.read_hysplit(filename)
    assert 'lat' in ds.variables.keys()
    assert 'lon' in ds.variables.keys()
    assert 'alt' in ds.variables.keys()
    assert 'PRESSURE' in ds.variables.keys()
    assert ds.dims["num_grids"] == 8
    assert ds.dims["num_trajectories"] == 1
    assert ds.dims['time'] == 121
    assert ds['age'].min() == -120



"""
This module contains procedures for reading and writing various ARM datasets.

"""

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=['arm', 'text', 'icartt', 'mpl', 'neon', 'noaagml', 'noaapsl', 'pysp2', 'hysplit'],
    submod_attrs={
        'arm': [
            'WriteDataset',
            'check_arm_standards',
            'create_ds_from_arm_dod',
            'read_arm_netcdf',
            'check_if_tar_gz_file',
            'read_arm_mmcr',
        ],
        'text': ['read_csv'],
        'icartt': ['read_icartt'],
        'mpl': ['proc_sigma_mplv5_read', 'read_sigma_mplv5'],
        'neon': ['read_neon_csv'],
        'noaagml': [
            'read_gml',
            'read_gml_co2',
            'read_gml_halo',
            'read_gml_met',
            'read_gml_ozone',
            'read_gml_radiation',
            'read_surfrad',
        ],
        'noaapsl': [
            'read_psl_wind_profiler',
            'read_psl_wind_profiler_temperature',
            'read_psl_parsivel',
            'read_psl_radar_fmcw_moment',
            'read_psl_surface_met',
        ],
        'pysp2': ['read_hk_file', 'read_sp2', 'read_sp2_dat'],
        'sodar': ['read_mfas_sodar'],
        'hysplit': ['read_hysplit'],
    },
)



import tempfile
from pathlib import Path

import numpy as np

import act
from act.tests import sample_files


def test_read_arm_netcdf():
    ds = act.io.arm.read_arm_netcdf([act.tests.EXAMPLE_MET1])
    assert 'temp_mean' in ds.variables.keys()
    assert 'rh_mean' in ds.variables.keys()
    assert ds.attrs['_arm_standards_flag'] == (1 << 0)

    with np.testing.assert_raises(OSError):
        ds = act.io.arm.read_arm_netcdf([])

    ds = act.io.arm.read_arm_netcdf([], return_None=True)
    assert ds is None
    ds = act.io.arm.read_arm_netcdf(['./randomfile.nc'], return_None=True)
    assert ds is None

    ds = act.io.arm.read_arm_netcdf([act.tests.EXAMPLE_MET_TEST1])
    assert 'time' in ds

    ds = act.io.arm.read_arm_netcdf([act.tests.EXAMPLE_MET_TEST2])
    assert ds['time'].values[10].astype('datetime64[ms]') == np.datetime64(
        '2019-01-01T00:10:00', 'ms'
    )

    ds = act.io.arm.read_arm_netcdf(
        act.tests.EXAMPLE_MET1, use_base_time=True, drop_variables='time'
    )
    assert 'time' in ds
    assert np.issubdtype(ds['time'].dtype, np.datetime64)
    assert ds['time'].values[10].astype('datetime64[ms]') == np.datetime64(
        '2019-01-01T00:10:00', 'ms'
    )

    del ds


def test_keep_variables():
    var_names = [
        'temp_mean',
        'rh_mean',
        'wdir_vec_mean',
        'tbrg_precip_total_corr',
        'atmos_pressure',
        'wspd_vec_mean',
        'pwd_pw_code_inst',
        'pwd_pw_code_15min',
        'pwd_mean_vis_10min',
        'logger_temp',
        'pwd_precip_rate_mean_1min',
        'pwd_cumul_snow',
        'pwd_mean_vis_1min',
        'pwd_pw_code_1hr',
        'org_precip_rate_mean',
        'tbrg_precip_total',
        'pwd_cumul_rain',
    ]
    var_names = var_names + ['qc_' + ii for ii in var_names]
    drop_variables = act.io.arm.keep_variables_to_drop_variables(act.tests.EXAMPLE_MET1, var_names)

    expected_drop_variables = [
        'wdir_vec_std',
        'base_time',
        'alt',
        'qc_wspd_arith_mean',
        'pwd_err_code',
        'logger_volt',
        'temp_std',
        'lon',
        'qc_logger_volt',
        'time_offset',
        'wspd_arith_mean',
        'lat',
        'vapor_pressure_std',
        'vapor_pressure_mean',
        'rh_std',
        'qc_vapor_pressure_mean',
    ]
    assert drop_variables.sort() == expected_drop_variables.sort()

    ds = act.io.arm.read_arm_netcdf(act.tests.EXAMPLE_MET1, keep_variables='temp_mean')
    assert list(ds.data_vars) == ['temp_mean']
    del ds

    var_names = ['temp_mean', 'qc_temp_mean']
    ds = act.io.arm.read_arm_netcdf(
        act.tests.EXAMPLE_MET1, keep_variables=var_names, drop_variables='nonsense'
    )
    assert list(ds.data_vars).sort() == var_names.sort()
    del ds

    var_names = ['temp_mean', 'qc_temp_mean', 'alt', 'lat', 'lon']
    ds = act.io.arm.read_arm_netcdf(
        act.tests.EXAMPLE_MET_WILDCARD, keep_variables=var_names, drop_variables=['lon']
    )
    var_names = list(set(var_names) - {'lon'})
    assert list(ds.data_vars).sort() == var_names.sort()
    del ds

    filenames = list(Path(file) for file in act.tests.EXAMPLE_MET_WILDCARD)
    var_names = ['temp_mean', 'qc_temp_mean', 'alt', 'lat', 'lon']
    ds = act.io.arm.read_arm_netcdf(filenames, keep_variables=var_names)
    assert list(ds.data_vars).sort() == var_names.sort()
    del ds


def test_read_arm_netcdf_mfdataset():
    met_ds = act.io.arm.read_arm_netcdf(act.tests.EXAMPLE_MET_WILDCARD)
    met_ds.load()
    assert 'temp_mean' in met_ds.variables.keys()
    assert 'rh_mean' in met_ds.variables.keys()
    assert len(met_ds.attrs['_file_times']) == 7
    assert met_ds.attrs['_arm_standards_flag'] == (1 << 0)
    met_ds.close()
    del met_ds

    met_ds = act.io.arm.read_arm_netcdf(act.tests.EXAMPLE_MET_WILDCARD, cleanup_qc=True)
    met_ds.load()
    var_name = 'temp_mean'
    qc_var_name = 'qc_' + var_name
    attr_names = [
        'long_name',
        'units',
        'flag_masks',
        'flag_meanings',
        'flag_assessments',
        'fail_min',
        'fail_max',
        'fail_delta',
        'standard_name',
    ]
    assert var_name in met_ds.variables.keys()
    assert qc_var_name in met_ds.variables.keys()
    assert sorted(attr_names) == sorted(list(met_ds[qc_var_name].attrs.keys()))
    assert met_ds[qc_var_name].attrs['flag_masks'] == [1, 2, 4, 8]
    assert met_ds[qc_var_name].attrs['flag_assessments'] == ['Bad', 'Bad', 'Bad', 'Indeterminate']
    met_ds.close()
    del met_ds


def test_io_dod():
    dims = {'time': 1440, 'drop_diameter': 50}

    try:
        ds = act.io.arm.create_ds_from_arm_dod(
            'vdis.b1', dims, version='1.2', scalar_fill_dim='time'
        )
        assert 'moment1' in ds
        assert len(ds['base_time'].values) == 1440
        assert len(ds['drop_diameter'].values) == 50
        with np.testing.assert_warns(UserWarning):
            ds2 = act.io.arm.create_ds_from_arm_dod('vdis.b1', dims, scalar_fill_dim='time')
        assert 'moment1' in ds2
        assert len(ds2['base_time'].values) == 1440
        assert len(ds2['drop_diameter'].values) == 50
        with np.testing.assert_raises(ValueError):
            ds = act.io.arm.create_ds_from_arm_dod('vdis.b1', {}, version='1.2')
        ds = act.io.arm.create_ds_from_arm_dod(
            sample_files.EXAMPLE_DOD, dims, version=1.2, scalar_fill_dim='time', local_file=True
        )
        assert 'moment1' in ds
        assert len(ds['base_time'].values) == 1440
        assert len(ds['drop_diameter'].values) == 50
    except Exception:
        return
    ds.close()
    ds2.close()


def test_io_write():
    sonde_ds = act.io.arm.read_arm_netcdf(sample_files.EXAMPLE_SONDE1)
    sonde_ds.clean.cleanup()

    with tempfile.TemporaryDirectory() as tmpdirname:
        write_file = Path(tmpdirname, Path(sample_files.EXAMPLE_SONDE1).name)
        keep_vars = ['tdry', 'qc_tdry', 'dp', 'qc_dp']
        for var_name in list(sonde_ds.data_vars):
            if var_name not in keep_vars:
                del sonde_ds[var_name]
        sonde_ds.write.write_netcdf(path=write_file, FillValue=-9999)

        sonde_ds_read = act.io.arm.read_arm_netcdf(str(write_file))
        assert list(sonde_ds_read.data_vars) == keep_vars
        assert isinstance(sonde_ds_read['qc_tdry'].attrs['flag_meanings'], str)
        assert sonde_ds_read['qc_tdry'].attrs['flag_meanings'].count('__') == 21
        for attr in ['qc_standards_version', 'qc_method', 'qc_comment']:
            assert attr not in list(sonde_ds_read.attrs)
        sonde_ds_read.close()
        del sonde_ds_read

    sonde_ds.close()

    sonde_ds = act.io.arm.read_arm_netcdf(sample_files.EXAMPLE_EBBR1)
    sonde_ds.clean.cleanup()
    assert 'fail_min' in sonde_ds['qc_home_signal_15'].attrs
    assert 'standard_name' in sonde_ds['qc_home_signal_15'].attrs
    assert 'flag_masks' in sonde_ds['qc_home_signal_15'].attrs

    with tempfile.TemporaryDirectory() as tmpdirname:
        cf_convention = 'CF-1.8'
        write_file = Path(tmpdirname, Path(sample_files.EXAMPLE_EBBR1).name)
        sonde_ds.write.write_netcdf(
            path=write_file,
            make_copy=False,
            join_char='_',
            cf_compliant=True,
            cf_convention=cf_convention,
        )

        sonde_ds_read = act.io.arm.read_arm_netcdf(str(write_file))

        assert cf_convention in sonde_ds_read.attrs['Conventions'].split()
        assert sonde_ds_read.attrs['FeatureType'] == 'timeSeries'
        global_att_keys = [ii for ii in sonde_ds_read.attrs.keys() if not ii.startswith('_')]
        assert global_att_keys[-1] == 'history'
        assert sonde_ds_read['alt'].attrs['axis'] == 'Z'
        assert sonde_ds_read['alt'].attrs['positive'] == 'up'

        sonde_ds_read.close()
        del sonde_ds_read

    sonde_ds.close()

    ds = act.io.arm.read_arm_netcdf(sample_files.EXAMPLE_CEIL1)
    with tempfile.TemporaryDirectory() as tmpdirname:
        cf_convention = 'CF-1.8'
        write_file = Path(tmpdirname, Path(sample_files.EXAMPLE_CEIL1).name)
        ds.write.write_netcdf(
            path=write_file,
            make_copy=False,
            join_char='_',
            cf_compliant=True,
            cf_convention=cf_convention,
        )

        ds_read = act.io.arm.read_arm_netcdf(str(write_file))

        assert cf_convention in ds_read.attrs['Conventions'].split()
        assert ds_read.attrs['FeatureType'] == 'timeSeriesProfile'
        assert len(ds_read.dims) > 1

        ds_read.close()
        del ds_read


def test_clean_cf_qc():
    with tempfile.TemporaryDirectory() as tmpdirname:
        ds = act.io.arm.read_arm_netcdf(sample_files.EXAMPLE_MET1, cleanup_qc=True)
        ds.load()
        var_name = 'temp_mean'
        qc_var_name = 'qc_' + var_name
        ds.qcfilter.remove_test(var_name, test_number=4)
        ds.qcfilter.remove_test(var_name, test_number=3)
        ds.qcfilter.remove_test(var_name, test_number=2)
        ds[qc_var_name].attrs['flag_masks'] = ds[qc_var_name].attrs['flag_masks'][0]
        flag_meanings = ds[qc_var_name].attrs['flag_meanings'][0]
        ds[qc_var_name].attrs['flag_meanings'] = flag_meanings.replace(' ', '__')
        flag_meanings = ds[qc_var_name].attrs['flag_assessments'][0]
        ds[qc_var_name].attrs['flag_assessments'] = flag_meanings.replace(' ', '__')

        write_file = str(Path(tmpdirname, Path(sample_files.EXAMPLE_MET1).name))
        ds.write.write_netcdf(path=write_file, cf_compliant=True)
        ds.close()
        del ds

        read_ds = act.io.arm.read_arm_netcdf(write_file, cleanup_qc=True)
        read_ds.load()

        assert type(read_ds[qc_var_name].attrs['flag_masks']).__module__ == 'numpy'
        assert read_ds[qc_var_name].attrs['flag_masks'].size == 1
        assert read_ds[qc_var_name].attrs['flag_masks'][0] == 1
        assert isinstance(read_ds[qc_var_name].attrs['flag_meanings'], list)
        assert len(read_ds[qc_var_name].attrs['flag_meanings']) == 1
        assert isinstance(read_ds[qc_var_name].attrs['flag_assessments'], list)
        assert len(read_ds[qc_var_name].attrs['flag_assessments']) == 1
        assert read_ds[qc_var_name].attrs['flag_assessments'] == ['Bad']
        assert read_ds[qc_var_name].attrs['flag_meanings'] == ['Value is equal to missing_value.']

        read_ds.close()
        del read_ds


def test_read_mmcr():
    results = act.tests.EXAMPLE_MMCR
    ds = act.io.arm.read_arm_mmcr(results)
    assert 'MeanDopplerVelocity_PR' in ds
    assert 'SpectralWidth_BL' in ds
    np.testing.assert_almost_equal(ds['Reflectivity_GE'].mean(), -34.62, decimal=2)
    np.testing.assert_almost_equal(ds['MeanDopplerVelocity_Receiver1'].max(), 9.98, decimal=2)



import numpy as np
import xarray as xr

import act


def test_calculate_sirs_variable():
    sirs_ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_SIRS)
    met_ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_MET1)

    ds = act.retrievals.radiation.calculate_dsh_from_dsdh_sdn(sirs_ds)
    assert np.isclose(np.nansum(ds['derived_down_short_hemisp'].values), 61157.71, atol=0.1)

    ds = act.retrievals.radiation.calculate_irradiance_stats(
        ds,
        variable='derived_down_short_hemisp',
        variable2='down_short_hemisp',
        threshold=60,
    )

    assert np.isclose(np.nansum(ds['diff_derived_down_short_hemisp'].values), 1335.12, atol=0.1)
    assert np.isclose(np.nansum(ds['ratio_derived_down_short_hemisp'].values), 400.31, atol=0.1)

    ds = act.retrievals.radiation.calculate_net_radiation(ds, smooth=30)
    assert np.ceil(np.nansum(ds['net_radiation'].values)) == 21915
    assert np.ceil(np.nansum(ds['net_radiation_smoothed'].values)) == 22316

    ds = act.retrievals.radiation.calculate_longwave_radiation(
        ds,
        temperature_var='temp_mean',
        vapor_pressure_var='vapor_pressure_mean',
        met_ds=met_ds,
    )
    assert np.ceil(ds['monteith_clear'].values[25]) == 239
    assert np.ceil(ds['monteith_cloudy'].values[30]) == 318
    assert np.ceil(ds['prata_clear'].values[35]) == 234

    new_ds = xr.merge([sirs_ds, met_ds], compat='override')
    ds = act.retrievals.radiation.calculate_longwave_radiation(
        new_ds, temperature_var='temp_mean', vapor_pressure_var='vapor_pressure_mean'
    )
    assert np.ceil(ds['monteith_clear'].values[25]) == 239
    assert np.ceil(ds['monteith_cloudy'].values[30]) == 318
    assert np.ceil(ds['prata_clear'].values[35]) == 234

    sirs_ds.close()
    met_ds.close()



"""
Stores the class for WindRoseDisplay.

"""

import warnings

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

# Import Local Libs
from ..utils import datetime_utils as dt_utils
from .plot import Display


class WindRoseDisplay(Display):
    """
    A class for handing wind rose plots.

    This is inherited from the :func:`act.plotting.Display`
    class and has therefore has the same attributes as that class.
    See :func:`act.plotting.Display`
    for more information. There are no additional attributes or parameters
    to this class.

    Examples
    --------
    To create a WindRoseDisplay object, simply do:

    .. code-block :: python

        sonde_ds = act.io.arm.read_arm_netcdf('sonde_data.nc')
        WindDisplay = act.plotting.WindRoseDisplay(sonde_ds, figsize=(8,10))

    """

    def __init__(self, ds, subplot_shape=(1,), ds_name=None, **kwargs):
        super().__init__(ds, subplot_shape, ds_name, subplot_kw=dict(projection='polar'), **kwargs)

    def set_thetarng(self, trng=(0.0, 360.0), subplot_index=(0,)):
        """
        Sets the theta range of the wind rose plot.

        Parameters
        ----------
        trng : 2-tuple
            The range (in degrees).
        subplot_index : 2-tuple
            The index of the subplot to set the degree range of.

        """
        if self.axes is not None:
            self.axes[subplot_index].set_thetamin(trng[0])
            self.axes[subplot_index].set_thetamax(trng[1])
            self.trng = trng
        else:
            raise RuntimeError('Axes must be initialized before' + ' changing limits!')

    def set_rrng(self, rrng, subplot_index=(0,)):
        """
        Sets the range of the radius of the wind rose plot.

        Parameters
        ----------
        rrng : 2-tuple
            The range for the plot radius (in %).
        subplot_index : 2-tuple
            The index of the subplot to set the radius range of.

        """
        if self.axes is not None:
            self.axes[subplot_index].set_rmin(rrng[0])
            self.axes[subplot_index].set_rmax(rrng[1])
            self.rrng = rrng
        else:
            raise RuntimeError('Axes must be initialized before' + ' changing limits!')

    def plot(
        self,
        dir_field,
        spd_field,
        dsname=None,
        subplot_index=(0,),
        cmap=None,
        set_title=None,
        num_dirs=20,
        spd_bins=None,
        tick_interval=3,
        legend_loc=0,
        legend_bbox=None,
        legend_title=None,
        calm_threshold=1.0,
        **kwargs,
    ):
        """
        Makes the wind rose plot from the given dataset.

        Parameters
        ----------
        dir_field : str
            The name of the field representing the wind direction (in degrees).
        spd_field : str
            The name of the field representing the wind speed.
        dsname : str
            The name of the datastream to plot from. Set to None to
            let ACT automatically try to determine this.
        subplot_index : 2-tuple
            The index of the subplot to place the plot on.
        cmap : str or matplotlib colormap
            The name of the matplotlib colormap to use.
        set_title : str
            The title of the plot.
        num_dirs : int
            The number of directions to split the wind rose into.
        spd_bins : 1D array-like
            The bin boundaries to sort the wind speeds into.
        tick_interval : int
            The interval (in %) for the ticks on the radial axis.
        legend_loc : int
            Legend location using matplotlib legend code
        legend_bbox : tuple
            Legend bounding box coordinates
        legend_title : string
            Legend title
        calm_threshold : float
            Winds below this threshold are considered to be calm.
        **kwargs : keyword arguments
            Additional keyword arguments will be passed into :func:plt.bar

        Returns
        -------
        ax : matplotlib axis handle
            The matplotlib axis handle corresponding to the plot.

        """
        if dsname is None and len(self._ds.keys()) > 1:
            raise ValueError(
                'You must choose a datastream when there are 2 '
                'or more datasets in the TimeSeriesDisplay '
                'object.'
            )
        elif dsname is None:
            dsname = list(self._ds.keys())[0]

        # Get data and dimensions
        dir_data = self._ds[dsname][dir_field].values
        spd_data = self._ds[dsname][spd_field].values

        # Get the current plotting axis, add day/night background and plot data
        if self.fig is None:
            self.fig = plt.figure()

        if self.axes is None:
            self.axes = np.array([plt.axes(projection='polar')])
            self.fig.add_axes(self.axes[0])

        if spd_bins is None:
            spd_bins = np.linspace(0, np.nanmax(spd_data), 10)

        # Make the bins so that 0 degrees N is in the center of the first bin
        # We need to wrap around

        deg_width = 360.0 / num_dirs
        dir_bins_mid = np.linspace(0.0, 360.0 - 3 * deg_width / 2.0, num_dirs)
        wind_hist = np.zeros((num_dirs, len(spd_bins) - 1))

        for i in range(num_dirs):
            if i == 0:
                with warnings.catch_warnings():
                    warnings.filterwarnings('ignore', 'invalid value encountered in.*')
                    the_range = np.logical_or(
                        dir_data < deg_width / 2.0, dir_data > 360.0 - deg_width / 2.0
                    )
            else:
                with warnings.catch_warnings():
                    warnings.filterwarnings('ignore', 'invalid value encountered in.*')
                    the_range = np.logical_and(
                        dir_data >= dir_bins_mid[i] - deg_width / 2,
                        dir_data <= dir_bins_mid[i] + deg_width / 2,
                    )
            hist, bins = np.histogram(spd_data[the_range], spd_bins)
            wind_hist[i] = hist

        wind_hist = wind_hist / np.sum(wind_hist) * 100
        mins = np.deg2rad(dir_bins_mid)

        # Do the first level
        if 'units' in self._ds[dsname][spd_field].attrs.keys():
            units = self._ds[dsname][spd_field].attrs['units']
        else:
            units = ''
        the_label = '%3.1f' % spd_bins[0] + '-' + '%3.1f' % spd_bins[1] + ' ' + units
        our_cmap = matplotlib.colormaps.get_cmap(cmap)
        our_colors = our_cmap(np.linspace(0, 1, len(spd_bins)))

        ax = self.axes[subplot_index]

        bars = [
            ax.bar(
                mins,
                wind_hist[:, 0],
                bottom=0,
                label=the_label,
                width=0.8 * np.deg2rad(deg_width),
                color=our_colors[0],
                **kwargs,
            )
        ]
        for i in range(1, len(spd_bins) - 1):
            the_label = '%3.1f' % spd_bins[i] + '-' + '%3.1f' % spd_bins[i + 1] + ' ' + units
            # Changing the bottom to be a sum of the previous speeds so that
            # it positions it correctly - Adam Theisen
            bars.append(
                ax.bar(
                    mins,
                    wind_hist[:, i],
                    label=the_label,
                    bottom=np.sum(wind_hist[:, :i], axis=1),
                    width=0.8 * np.deg2rad(deg_width),
                    color=our_colors[i],
                    **kwargs,
                )
            )
        ax.legend(loc=legend_loc, bbox_to_anchor=legend_bbox, title=legend_title)
        ax.set_theta_zero_location('N')
        ax.set_theta_direction(-1)

        # Add an annulus with text stating % of time calm
        pct_calm = np.sum(spd_data <= calm_threshold) / len(spd_data) * 100
        ax.set_rorigin(-2.5)
        ax.annotate('%3.2f%%\n calm' % pct_calm, xy=(0, -2.5), ha='center', va='center')

        # Set the ticks to be nice numbers
        tick_max = tick_interval * round(np.nanmax(np.cumsum(wind_hist, axis=1)) / tick_interval)
        rticks = np.arange(0, tick_max, tick_interval)
        rticklabels = [('%d' % x + '%') for x in rticks]
        ax.set_rticks(rticks)
        ax.set_yticklabels(rticklabels)

        # Set Title
        if set_title is None:
            set_title = ' '.join(
                [
                    dsname,
                    'on',
                    dt_utils.numpy_to_arm_date(self._ds[dsname].time.values[0]),
                ]
            )
        ax.set_title(set_title)

        self.axes[subplot_index] = ax

        return ax

    def plot_data(
        self,
        dir_field,
        spd_field,
        data_field,
        dsname=None,
        subplot_index=(0,),
        plot_type='Line',
        line_color=None,
        set_title=None,
        num_dirs=30,
        num_data_bins=30,
        calm_threshold=1.0,
        line_plot_calc='mean',
        clevels=30,
        contour_type='count',
        cmap=None,
        **kwargs,
    ):
        """
        Makes a data rose plot in line or boxplot form from the given data.

        Parameters
        ----------
        dir_field : str
            The name of the field representing the wind direction (in degrees).
        spd_field : str
            The name of the field representing the wind speed.
        data_field : str
            Name of the field to plot.  Default is to plot mean values.
        dsname : str
            The name of the datastream to plot from. Set to None to
            let ACT automatically try to determine this.
        subplot_index : 2-tuple
            The index of the subplot to place the plot on.
        plot_type : str
            Type of plot to create.  Defaults to a line plot but the full options include
            'line', 'contour', and 'boxplot'
        line_color : str
            Color to use for the line
        set_title : str
            The title of the plot.
        num_dirs : int
            The number of directions to split the wind rose into.
        num_data_bins : int
            The number of bins to use for data processing if doing a contour plot
        calm_threshold : float
            Winds below this threshold are considered to be calm.
        line_plot_calc : str
            What values to display for the line plot.  Defaults to 'mean',
            but other options are 'median' and 'stdev'
        clevels : int
            Number of contour levels to plot
        contour_type : str
            Type of contour plot to do.  Default is 'count' which displays a
            heatmap of where values are occuring most along with wind directions
            The other option is 'mean' which will do a wind direction x wind speed
            plot with the contours of the mean values for each wind dir/speed.
            num_data_bins will be used for number of wind speed bins
        cmap : str or matplotlib colormap
            The name of the matplotlib colormap to use.
        **kwargs : keyword arguments
            Additional keyword arguments will be passed into :func:plt.bar

        Returns
        -------
        ax : matplotlib axis handle
            The matplotlib axis handle corresponding to the plot.

        """
        if dsname is None and len(self._ds.keys()) > 1:
            raise ValueError(
                'You must choose a datastream when there are 2 '
                'or more datasets in the TimeSeriesDisplay '
                'object.'
            )
        elif dsname is None:
            dsname = list(self._ds.keys())[0]

        # Get data and dimensions
        # Throw out calm winds for the analysis
        ds = self._ds[dsname]
        ds = ds.where(ds[spd_field] >= calm_threshold)
        dir_data = ds[dir_field].values
        data = ds[data_field].values

        # Set the bins
        dir_bins_mid = np.linspace(0.0, 360.0, num_dirs + 1)

        # Run through the data and bin based on the wind direction and plot type
        arr = []
        bins = []
        for i, d in enumerate(dir_bins_mid):
            if i < len(dir_bins_mid) - 1:
                idx = np.where((dir_data > d) & (dir_data <= dir_bins_mid[i + 1]))[0]
                bins.append(d + (dir_bins_mid[i + 1] - d) / 2.0)
            else:
                idx = np.where((dir_data > d) & (dir_data <= 360.0))[0]
                bins.append(d + (360.0 - d) / 2.0)

            if plot_type == 'line':
                if line_plot_calc == 'mean':
                    arr.append(np.nanmean(data[idx]))
                    plot_type_str = 'Mean of'
                elif line_plot_calc == 'median':
                    arr.append(np.nanmedian(data[idx]))
                    plot_type_str = 'Median of'
                elif line_plot_calc == 'stdev':
                    plot_type_str = 'Standard Deviation of'
                    arr.append(np.nanstd(data[idx]))
                else:
                    raise ValueError('Please pick an available option')
            elif plot_type == 'boxplot':
                arr.append(data[idx])

        # Plot data for each plot type
        if plot_type == 'line':
            # Add the first values to the end of the array to have a
            # complete circle
            bins.append(bins[0])
            arr.append(arr[0])
            self.axes[subplot_index].plot(np.deg2rad(bins), arr, **kwargs)
        elif plot_type == 'boxplot':
            # Plot boxplot
            self.axes[subplot_index].boxplot(
                arr, positions=np.deg2rad(bins), showmeans=False, **kwargs
            )
            if bins[-1] == 360:
                bins[-1] = 0
            self.axes[subplot_index].xaxis.set_ticklabels(np.ceil(bins))
            plot_type_str = 'Boxplot of'
        elif plot_type == 'contour':
            # Calculate a histogram to plot out a contour for
            if contour_type == 'count':
                idx = np.where((~np.isnan(dir_data)) & (~np.isnan(data)))[0]
                hist, xedges, yedges = np.histogram2d(
                    dir_data[idx], data[idx], bins=[num_dirs, num_data_bins]
                )
                hist = np.insert(hist, -1, hist[0], axis=0)
                cplot = self.axes[subplot_index].contourf(
                    np.deg2rad(xedges),
                    yedges[0:-1],
                    np.transpose(hist),
                    cmap=cmap,
                    levels=clevels,
                    **kwargs,
                )
                plot_type_str = 'Heatmap of'
                cbar = self.fig.colorbar(cplot, ax=self.axes[subplot_index])
                cbar.ax.set_ylabel('Count')
            elif contour_type == 'mean':
                # Produce direction (x-axis) and speed (y-axis) plots displaying the mean
                # as the contours.
                spd_data = ds[spd_field].values
                spd_bins = np.linspace(0, ds[spd_field].max(), num_data_bins + 1)
                spd_bins = np.insert(spd_bins, 1, calm_threshold)
                #  Set up an array and cycle through the data, binning them by speed/direction
                mean_data = np.zeros([len(bins), len(spd_bins)])
                for i in range(len(bins) - 1):
                    for j in range(len(spd_bins)):
                        if j < len(spd_bins) - 1:
                            idx = np.where(
                                (spd_data >= spd_bins[j])
                                & (spd_data < spd_bins[j + 1])
                                & (dir_data >= bins[i])
                                & (dir_data < bins[i + 1])
                            )[0]
                        else:
                            idx = np.where(
                                (spd_data >= spd_bins[j])
                                & (dir_data >= bins[i])
                                & (dir_data < bins[i + 1])
                            )[0]
                        mean_data[i, j] = np.nanmean(data[idx])

                # Necessary to produce the full polar contour without having gaps
                mean_data = np.insert(mean_data, -1, mean_data[0, :], axis=0)
                bins.append(bins[0])
                mean_data[-1, :] = mean_data[0, :]

                # In order to properly handle vmin/vmax in contours, need to adjust
                # the levels plotted and remove the keywords to contourf
                vmin = np.nanmin(mean_data)
                vmax = np.nanmax(mean_data)
                if 'vmin' in kwargs:
                    vmin = kwargs.get('vmin')
                    kwargs.pop('vmin', None)
                if 'vmax' in kwargs:
                    vmax = kwargs.get('vmax')
                    kwargs.pop('vmax', None)

                clevels = np.linspace(vmin, vmax, clevels)
                cplot = self.axes[subplot_index].contourf(
                    np.deg2rad(bins),
                    spd_bins,
                    np.transpose(mean_data),
                    cmap=cmap,
                    levels=clevels,
                    extend='both',
                    **kwargs,
                )
                plot_type_str = 'Mean of'
                cbar = self.fig.colorbar(cplot, ax=self.axes[subplot_index])
                cbar.ax.set_ylabel('Mean')
        else:
            raise ValueError('Please choose an available plot type')

        # Set axis parameters so that it's a standard wind rose style
        self.axes[subplot_index].set_theta_zero_location('N')
        self.axes[subplot_index].set_theta_direction(-1)

        # Set Title
        sdate = (dt_utils.numpy_to_arm_date(self._ds[dsname].time.values[0]),)
        edate = (dt_utils.numpy_to_arm_date(self._ds[dsname].time.values[-1]),)

        if sdate == edate:
            date_str = 'on ' + sdate[0]
        else:
            date_str = 'from ' + sdate[0] + ' to ' + edate[0]
        if 'units' in ds[data_field].attrs:
            units = ds[data_field].attrs['units']
        else:
            units = ''
        if set_title is None:
            set_title = ' '.join(
                [plot_type_str, data_field + ' (' + units + ')', 'by\n', dir_field, date_str]
            )
        self.axes[subplot_index].set_title(set_title)
        plt.tight_layout(h_pad=1.05)

        return self.axes[subplot_index]



import numpy as np
import pytest

import act


@pytest.mark.skipif(not act.io.icartt._ICARTT_AVAILABLE, reason='ICARTT is not installed.')
def test_read_icartt():
    result = act.io.icartt.read_icartt(act.tests.EXAMPLE_AAF_ICARTT)
    assert 'pitch' in result
    assert len(result['time'].values) == 14087
    assert result['true_airspeed'].units == 'm/s'
    assert 'Revision' in result.attrs
    np.testing.assert_almost_equal(result['static_pressure'].mean(), 708.75, decimal=2)



from datetime import datetime

import numpy as np
import pandas as pd

import act


def test_dates_between():
    start_date = '20191201'
    end_date = '20201201'
    date_list = act.utils.dates_between(start_date, end_date)
    start_string = datetime.strptime(start_date, '%Y%m%d').strftime('%Y-%m-%d')
    end_string = datetime.strptime(end_date, '%Y%m%d').strftime('%Y-%m-%d')
    answer = np.arange(start_string, end_string, dtype='datetime64[D]')
    answer = np.append(answer, answer[-1] + 1)
    answer = answer.astype('datetime64[s]').astype(int)
    answer = [datetime.utcfromtimestamp(ii) for ii in answer]

    assert date_list == answer


def test_datetime64_to_datetime():
    time_datetime = [
        datetime(2019, 1, 1, 1, 0),
        datetime(2019, 1, 1, 1, 1),
        datetime(2019, 1, 1, 1, 2),
        datetime(2019, 1, 1, 1, 3),
        datetime(2019, 1, 1, 1, 4),
    ]

    time_datetime64 = [
        np.datetime64(datetime(2019, 1, 1, 1, 0)),
        np.datetime64(datetime(2019, 1, 1, 1, 1)),
        np.datetime64(datetime(2019, 1, 1, 1, 2)),
        np.datetime64(datetime(2019, 1, 1, 1, 3)),
        np.datetime64(datetime(2019, 1, 1, 1, 4)),
    ]

    time_datetime64_to_datetime = act.utils.datetime_utils.datetime64_to_datetime(time_datetime64)
    assert time_datetime == time_datetime64_to_datetime


def test_reduce_time_ranges():
    time = pd.date_range(start='2020-01-01T00:00:00', freq='1min', periods=100)
    time = time.to_list()
    time = time[0:50] + time[60:]
    result = act.utils.datetime_utils.reduce_time_ranges(time)
    assert len(result) == 2
    assert result[1][1].minute == 39

    result = act.utils.datetime_utils.reduce_time_ranges(time, broken_barh=True)
    assert len(result) == 2


def test_date_parser():
    datestring = '20111001'
    output_format = '%Y/%m/%d'

    test_string = act.utils.date_parser(datestring, output_format, return_datetime=False)
    assert test_string == '2011/10/01'

    test_datetime = act.utils.date_parser(datestring, output_format, return_datetime=True)
    assert test_datetime == datetime(2011, 10, 1)


def test_date_parser_minute_second():
    date_string = '2020-01-01T12:00:00'
    parsed_date = act.utils.date_parser(date_string, return_datetime=True)
    assert parsed_date == datetime(2020, 1, 1, 12, 0, 0)

    output_format = parsed_date.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z'
    assert output_format == '2020-01-01T12:00:00.000Z'


def test_adjust_timestamp():
    file = act.tests.sample_files.EXAMPLE_EBBR1
    ds = act.io.arm.read_arm_netcdf(file)
    ds = act.utils.datetime_utils.adjust_timestamp(ds)
    assert ds['time'].values[0] == np.datetime64('2019-11-24T23:30:00.000000000')

    ds = act.utils.datetime_utils.adjust_timestamp(ds, offset=-60 * 60)
    assert ds['time'].values[0] == np.datetime64('2019-11-24T22:30:00.000000000')



import numpy as np
import pytest

import act
from act.io import read_psl_surface_met, read_psl_wind_profiler_temperature
from act.tests import sample_files


def test_read_psl_wind_profiler():
    test_ds_low, test_ds_hi = act.io.noaapsl.read_psl_wind_profiler(
        act.tests.EXAMPLE_NOAA_PSL, transpose=False
    )
    # test dimensions
    assert 'time' and 'HT' in test_ds_low.dims.keys()
    assert 'time' and 'HT' in test_ds_hi.dims.keys()
    assert test_ds_low.dims['time'] == 4
    assert test_ds_hi.dims['time'] == 4
    assert test_ds_low.dims['HT'] == 49
    assert test_ds_hi.dims['HT'] == 50

    # test coordinates
    assert (test_ds_low.coords['HT'][0:5] == np.array([0.151, 0.254, 0.356, 0.458, 0.561])).all()
    assert (
        test_ds_low.coords['time'][0:2]
        == np.array(
            ['2021-05-05T15:00:01.000000000', '2021-05-05T15:15:49.000000000'],
            dtype='datetime64[ns]',
        )
    ).all()

    # test attributes
    assert test_ds_low.attrs['site_identifier'] == 'CTD'
    assert test_ds_low.attrs['data_type'] == 'WINDS'
    assert test_ds_low.attrs['revision_number'] == '5.1'
    assert test_ds_low.attrs['latitude'] == 34.66
    assert test_ds_low.attrs['longitude'] == -87.35
    assert test_ds_low.attrs['elevation'] == 187.0
    assert (
        test_ds_low.attrs['beam_azimuth'] == np.array([38.0, 38.0, 308.0], dtype='float32')
    ).all()
    assert (
        test_ds_low.attrs['beam_elevation'] == np.array([90.0, 74.7, 74.7], dtype='float32')
    ).all()
    assert test_ds_low.attrs['consensus_average_time'] == 24
    assert test_ds_low.attrs['oblique-beam_vertical_correction'] == 0
    assert test_ds_low.attrs['number_of_beams'] == 3
    assert test_ds_low.attrs['number_of_range_gates'] == 49
    assert test_ds_low.attrs['number_of_gates_oblique'] == 49
    assert test_ds_low.attrs['number_of_gates_vertical'] == 49
    assert test_ds_low.attrs['number_spectral_averages_oblique'] == 50
    assert test_ds_low.attrs['number_spectral_averages_vertical'] == 50
    assert test_ds_low.attrs['pulse_width_oblique'] == 708
    assert test_ds_low.attrs['pulse_width_vertical'] == 708
    assert test_ds_low.attrs['inner_pulse_period_oblique'] == 50
    assert test_ds_low.attrs['inner_pulse_period_vertical'] == 50
    assert test_ds_low.attrs['full_scale_doppler_value_oblique'] == 20.9
    assert test_ds_low.attrs['full_scale_doppler_value_vertical'] == 20.9
    assert test_ds_low.attrs['delay_to_first_gate_oblique'] == 4000
    assert test_ds_low.attrs['delay_to_first_gate_vertical'] == 4000
    assert test_ds_low.attrs['spacing_of_gates_oblique'] == 708
    assert test_ds_low.attrs['spacing_of_gates_vertical'] == 708

    # test fields
    assert test_ds_low['RAD1'].shape == (4, 49)
    assert test_ds_hi['RAD1'].shape == (4, 50)
    assert (test_ds_low['RAD1'][0, 0:5] == np.array([0.2, 0.1, 0.1, 0.0, -0.1])).all()
    assert (test_ds_hi['RAD1'][0, 0:5] == np.array([0.1, 0.1, -0.1, 0.0, -0.2])).all()

    assert test_ds_low['SPD'].shape == (4, 49)
    assert test_ds_hi['SPD'].shape == (4, 50)
    assert (test_ds_low['SPD'][0, 0:5] == np.array([2.5, 3.3, 4.3, 4.3, 4.8])).all()
    assert (test_ds_hi['SPD'][0, 0:5] == np.array([3.7, 4.6, 6.3, 5.2, 6.8])).all()

    # test transpose
    test_ds_low, test_ds_hi = act.io.noaapsl.read_psl_wind_profiler(
        act.tests.EXAMPLE_NOAA_PSL, transpose=True
    )
    assert test_ds_low['RAD1'].shape == (49, 4)
    assert test_ds_hi['RAD1'].shape == (50, 4)
    assert test_ds_low['SPD'].shape == (49, 4)
    assert test_ds_hi['SPD'].shape == (50, 4)
    test_ds_low.close()


def test_read_psl_wind_profiler_temperature():
    ds = read_psl_wind_profiler_temperature(act.tests.EXAMPLE_NOAA_PSL_TEMPERATURE)

    assert ds.attrs['site_identifier'] == 'CTD'
    assert ds.attrs['elevation'] == 600.0
    assert ds.T.values[0] == 33.2


def test_read_psl_surface_met():
    ds = read_psl_surface_met(sample_files.EXAMPLE_NOAA_PSL_SURFACEMET)
    assert ds.time.size == 2
    assert np.isclose(np.sum(ds['Pressure'].values), 1446.9)
    assert np.isclose(ds['lat'].values, 38.972425)
    assert ds['lat'].attrs['units'] == 'degree_N'
    assert ds['Upward_Longwave_Irradiance'].attrs['long_name'] == 'Upward Longwave Irradiance'
    assert ds['Upward_Longwave_Irradiance'].dtype.str == '<f4'

    with pytest.raises(Exception):
        ds = read_psl_surface_met('aaa22001.00m')


def test_read_psl_parsivel():
    url = [
        'https://downloads.psl.noaa.gov/psd2/data/realtime/DisdrometerParsivel/Stats/ctd/2022/002/ctd2200200_stats.txt',
        'https://downloads.psl.noaa.gov/psd2/data/realtime/DisdrometerParsivel/Stats/ctd/2022/002/ctd2200201_stats.txt',
        'https://downloads.psl.noaa.gov/psd2/data/realtime/DisdrometerParsivel/Stats/ctd/2022/002/ctd2200202_stats.txt',
    ]

    ds = act.io.noaapsl.read_psl_parsivel(url)
    assert 'number_density_drops' in ds
    assert np.max(ds['number_density_drops'].values) == 355
    assert ds['number_density_drops'].values[10, 10] == 201

    ds = act.io.noaapsl.read_psl_parsivel(
        'https://downloads.psl.noaa.gov/psd2/data/realtime/DisdrometerParsivel/Stats/ctd/2022/002/ctd2200201_stats.txt'
    )
    assert 'number_density_drops' in ds


def test_read_psl_fmcw_moment():
    result = act.discovery.download_noaa_psl_data(
        site='kps', instrument='Radar FMCW Moment', startdate='20220815', hour='06'
    )
    ds = act.io.noaapsl.read_psl_radar_fmcw_moment([result[-1]])
    assert 'range' in ds
    np.testing.assert_almost_equal(ds['reflectivity_uncalibrated'].mean(), 2.37, decimal=2)
    assert ds['range'].max() == 10040.0
    assert len(ds['time'].values) == 115


def test_read_psl_sband_moment():
    result = act.discovery.download_noaa_psl_data(
        site='ctd', instrument='Radar S-band Moment', startdate='20211225', hour='06'
    )
    ds = act.io.noaapsl.read_psl_radar_sband_moment([result[-1]])
    assert 'range' in ds
    np.testing.assert_almost_equal(ds['reflectivity_uncalibrated'].mean(), 1.00, decimal=2)
    assert ds['range'].max() == 9997.0
    assert len(ds['time'].values) == 37



import tempfile
from pathlib import Path

import act


def test_calculate_dqr_times():
    ebbr1_ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_EBBR1)
    ebbr2_ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_EBBR2)
    brs_ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_BRS)
    ebbr1_result = act.utils.calculate_dqr_times(ebbr1_ds, variable=['soil_temp_1'], threshold=2)
    ebbr2_result = act.utils.calculate_dqr_times(
        ebbr2_ds, variable=['rh_bottom_fraction'], qc_bit=3, threshold=2
    )
    ebbr3_result = act.utils.calculate_dqr_times(
        ebbr2_ds, variable=['rh_bottom_fraction'], qc_bit=3
    )
    brs_result = act.utils.calculate_dqr_times(
        brs_ds, variable='down_short_hemisp_min', qc_bit=2, threshold=30
    )
    assert ebbr1_result == [('2019-11-25 02:00:00', '2019-11-25 04:30:00')]
    assert ebbr2_result == [('2019-11-30 00:00:00', '2019-11-30 11:00:00')]
    assert brs_result == [('2019-07-05 01:57:00', '2019-07-05 11:07:00')]
    assert ebbr3_result is None
    with tempfile.TemporaryDirectory() as tmpdirname:
        write_file = Path(tmpdirname)
        brs_result = act.utils.calculate_dqr_times(
            brs_ds,
            variable='down_short_hemisp_min',
            qc_bit=2,
            threshold=30,
            txt_path=str(write_file),
        )

    brs_result = act.utils.calculate_dqr_times(
        brs_ds,
        variable='down_short_hemisp_min',
        qc_bit=2,
        threshold=30,
        return_missing=False,
    )
    assert len(brs_result[0]) == 2

    ebbr1_ds.close()
    ebbr2_ds.close()
    brs_ds.close()
